package fr.hugol;

import org.graphstream.algorithm.Toolkit;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.*;

public class VirusSimulation {
    public static void main(String[] args) {
        Graph graph = Main.createBarabasiAlbert(); // Crée un graphe Barabasi-Albert
        int days = 90;
        int runs = 20;
        double beta = 0.143; // Taux d'infection
        double mu = 0.067;  // Taux de récupération

        simulate(graph, beta, mu, days, runs, 1); // Scénario 1 : Aucun contrôle
        simulate(graph, beta, mu, days, runs, 2); // Scénario 2 : Immunisation aléatoire
        simulate(graph, beta, mu, days, runs, 3); // Scénario 3 : Immunisation sélective
    }

    public static void simulate(Graph graph, double beta, double mu, int days, int runs, int scenario) {
        System.out.println("Scenario " + scenario);
        List<double[]> allRunsFractions = new ArrayList<>();

        for (int run = 0; run < runs; run++) {
            initializeGraph(graph, scenario);
            Map<Node, String> states = initializeStates(graph);

            double[] dailyFractions = new double[days];
            for (int day = 0; day < days; day++) {
                states = simulateDay(graph, states, beta, mu);
                dailyFractions[day] = calculateFractionInfected(states);
            }
            allRunsFractions.add(dailyFractions);
        }

        double[] avgFractions = averageFractions(allRunsFractions, days);
        saveFractionsToFile(avgFractions, "gnuplot/tp2/scenario_" + scenario + ".dat");


        if (scenario == 3 || scenario == 2) {
            calculateGroupDegrees(graph);
        }
    }

    private static void initializeGraph(Graph graph, int scenario) {
        graph.nodes().forEach(n -> n.removeAttribute("immune"));

        if (scenario == 2) { // Immunisation aléatoire
            for (Node node : Toolkit.randomNodeSet(graph, graph.getNodeCount() / 2)) {
                node.setAttribute("immune", true);
            }
        } else if (scenario == 3) { // Immunisation sélective
            for (Node node : Toolkit.randomNodeSet(graph, graph.getNodeCount() / 2)) {
                for (Node neighbor : node.neighborNodes().toList()) {
                    neighbor.setAttribute("immune", true);
                }
            }
        }
    }

    private static Map<Node, String> initializeStates(Graph graph) {
        Map<Node, String> states = new HashMap<>();
        for (Node node : graph) {
            if (node.hasAttribute("immune")) {
                states.put(node, "R");
            } else {
                states.put(node, "S"); // Susceptible
            }
        }

        // Infecter un patient zéro
        Node patientZero = Toolkit.randomNode(graph);
        if (!states.get(patientZero).equals("R")) {
            states.put(patientZero, "I");
        }
        return states;
    }

    private static Map<Node, String> simulateDay(Graph graph, Map<Node, String> states, double beta, double mu) {
        Map<Node, String> newStates = new HashMap<>(states);
        for (Node node : graph) {
            String state = states.get(node);
            if (state.equals("I")) { // Infecté
                for (Node neighbor : node.neighborNodes().toList()) {
                    if (states.get(neighbor).equals("S") && Math.random() < beta) {
                        newStates.put(neighbor, "I"); // Infecter
                    }
                }
                if (Math.random() < mu) {
                    newStates.put(node, "R"); // Guérir
                }
            }
        }
        return newStates;
    }

    private static double calculateFractionInfected(Map<Node, String> states) {
        long infectedCount = states.values().stream().filter(s -> s.equals("I")).count();
        long susceptibleCount = states.values().stream().filter(s -> s.equals("S")).count();
        return (double) infectedCount / (infectedCount + susceptibleCount);
    }

    private static double[] averageFractions(List<double[]> allRunsFractions, int days) {
        double[] averages = new double[days];
        for (int day = 0; day < days; day++) {
            double sum = 0;
            for (double[] run : allRunsFractions) {
                sum += run[day];
            }
            averages[day] = sum / allRunsFractions.size();
        }
        return averages;
    }

    private static void saveFractionsToFile(double[] fractions, String fileName) {
        try (PrintWriter writer = new PrintWriter(new FileWriter(fileName))) {
            for (int day = 0; day < fractions.length; day++) {
                writer.printf(Locale.US, "%d\t%.5f%n", day, fractions[day]);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private static void calculateGroupDegrees(Graph graph) {
        System.out.println("Calculating degrees for groups 0 and 1.");

        double group0TotalDegree = 0;
        int group0Count = 0;
        double group1TotalDegree = 0;
        int group1Count = 0;

        for (Node node : graph) {
            if (node.hasAttribute("immune")) {
                group1TotalDegree += node.getDegree();
                group1Count++;
            } else {
                group0TotalDegree += node.getDegree();
                group0Count++;
            }
        }

        double group0AvgDegree = group0Count > 0 ? group0TotalDegree / group0Count : 0;
        double group1AvgDegree = group1Count > 0 ? group1TotalDegree / group1Count : 0;

        System.out.printf("Group 0 average degree: %.2f\n", group0AvgDegree);
        System.out.printf("Group 1 average degree: %.2f\n", group1AvgDegree);
    }
}
